from datetime import datetime
import re

def create_final_result(dictionary):
    result = ""
    for key, value in dictionary.items():
        if isinstance(value, dict):
            sub_result = create_final_result(value)
            if sub_result == 'consider':
                return 'consider'
            elif sub_result == 'clear':
                result = 'clear'
        elif value in ['clear', 'consider', ""]:
            if value == 'consider':
                return 'consider'
            elif result != 'clear':
                result = value
    return result

def age_validation(dob, age_threshold=18):
    age_val = {
        "breakdown": {
            "minimum_accepted_age": {
            "properties": {},
            "result": ""
            }
        },
        "result": ""
        }
    
    dob_date = datetime.strptime(dob, "%d/%m/%Y")

    current_date = datetime.now()

    age = current_date.year - dob_date.year - ((current_date.month, current_date.day) < (dob_date.month, dob_date.day))

    if age>=age_threshold:
        age_val["breakdown"]["minimum_accepted_age"]["result"] = "clear"
        age_val["result"] = "clear"
    else:
        age_val["breakdown"]["minimum_accepted_age"]["result"] = "consider"
        age_val["result"] = "consider"

    return age_val
    
def created_at():
    current_datetime = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')

    return current_datetime

def identify_document_type(text):
    text = text.upper()
    emirates_id_pattern = r'\b(ILARE\w*|IDARE\w*|RESIDENT IDENTITY)\b'
    passport_pattern = r'\b(PASSPORT|PPT)\b'
    driver_license_pattern = r'\b(DRIVER|LICENSE|DL)\b'

    if re.search(emirates_id_pattern, text):
        return "EID"

    if re.search(passport_pattern, text):
        return "PASSPORT"

    if re.search(driver_license_pattern, text):
        return "DL"

    return "Unknown"

def identify_front_id(text):
    front_id_keywords = ['Resident Identity', 'ID Number']
    pattern = '|'.join(map(re.escape, front_id_keywords))
    
    if re.search(pattern, text, re.IGNORECASE):
        return True
    else:
        return False

def identify_back_id(text):
    back_id_keywords = ['ILARE', 'IDARE', 'Signature']
    pattern = '|'.join(map(re.escape, back_id_keywords))
    
    if re.search(pattern, text, re.IGNORECASE):
        return True
    else:
        return False

def data_comparison_check():
    ## sending default as clear - identified from current nym's table results
    data_comparison = {
      "breakdown": {
        "date_of_birth": {
          "properties": {}
        },
        "date_of_expiry": {
          "properties": {}
        },
        "document_numbers": {
          "properties": {}
        },
        "document_type": {
          "properties": {}
        },
        "first_name": {
          "properties": {}
        },
        "gender": {
          "properties": {}
        },
        "issuing_country": {
          "properties": {}
        },
        "last_name": {
          "properties": {}
        }
      }
    }

    return data_comparison

def data_consistency_check(data, front_id_text, back_id_text):
    data_consistency = {
      "breakdown": {
        "date_of_birth": {
          "properties": {},
          "result": "clear"
        },
        "date_of_expiry": {
          "properties": {},
          "result": "clear"
        },
        "document_numbers": {
          "properties": {},
          "result": "clear"
        },
        "document_type": {
          "properties": {},
          "result": "clear"
        },
        "first_name": {
          "properties": {},
          "result": "clear"
        },
        "gender": {
          "properties": {},
          "result": "clear"
        },
        "issuing_country": {
          "properties": {},
          "result": "clear"
        },
        "last_name": {
          "properties": {},
          "result": "clear"
        },
        "multiple_data_sources_present": {
          "properties": {},
          "result": "clear"
        },
        "nationality": {
          "properties": {},
          "result": "clear"
        }
      },
      "result": "clear"
    }

    #### For data consistency compare data from different sources, like id and passport. 
    #### so the dob from id should match with dob extracted from passport

    doc_type1 = identify_document_type(front_id_text)
    doc_type2 = identify_document_type(back_id_text)
    if doc_type1 == 'EID' or doc_type2=='EID':
        data_consistency['breakdown']['document_type']['result'] = 'clear'
    else:
        data_consistency['breakdown']['document_type']['result'] = 'consider'
        data_consistency['result'] = 'consider'
    
    return data_consistency

def data_validation_check(data):
    data_validation = {
    "breakdown": {
        "date_of_birth": {
            "properties": {},
            "result": "clear"
        },
        ## pending
        "document_expiration": {
            "properties": {},
            "result": "clear"
        },
        "document_numbers": {
            "properties": {},
            "result": "clear"
        },
        "expiry_date": {
            "properties": {},
            "result": "clear"
        },
        "gender": {
            "properties": {},
            "result": "clear"
        },
        "mrz": {
            "properties": {},
            "result": "clear"
        },
        "barcode": {
            "properties": {},
            "result": "clear"
        }
    },
    "result": "clear"
}

    try:
        dob = data.get('dob')
        parsed_date = datetime.strptime(dob, "%d/%m/%Y")
        data_validation["breakdown"]['date_of_birth']["result"] = 'clear'
    except ValueError:
        data_validation["breakdown"]['date_of_birth']["result"] = 'consider'

    doc_no = data.get('card_number')
    if len(doc_no)==9:
        data_validation["breakdown"]['document_numbers']["result"] = 'clear'
    else:
        data_validation["breakdown"]['document_numbers']["result"] = 'consider'

    try:
        doe = data.get('expiry_date')
        parsed_date = datetime.strptime(doe, "%d/%m/%Y")
        data_validation["breakdown"]['expiry_date']["result"] = 'clear'
    except ValueError:
        data_validation["breakdown"]['expiry_date']["result"] = 'consider'

    gender = data.get('gender')
    if gender.isalpha() and len(gender) == 1:
        data_validation["breakdown"]['gender']["result"] = 'clear'
    else:
        data_validation["breakdown"]['gender']["result"] = 'consider'
    
    mrz = data.get('mrz')
    mrz1 = data.get('mrz1')
    mrz2 = data.get('mrz2')
    mrz3 = data.get('mrz3')
    if len(mrz) == 1 and mrz1 and mrz2 and mrz3:
        data_validation["breakdown"]['mrz']["result"] = 'clear'
    else:
        data_validation["breakdown"]['mrz']["result"] = 'consider'
    
    result = create_final_result(data_validation)
    data_validation['result'] = result

    # if data_validation["breakdown"]['date_of_birth']["result"]=='clear' and data_validation["breakdown"]['expiry_date']["result"]=='clear' and data_validation["breakdown"]['gender']["result"]=='clear' and data_validation["breakdown"]['mrz']["result"]=='clear':
    #     data_validation['result'] = 'clear'

    return data_validation

## pending
def image_integrity_check(front_id_text, back_id_text, coloured, blurred, glare, missing_fields):
    image_integrity = {
      "breakdown": {
        "colour_picture": {
          "properties": {},
          "result": "clear"
        },
        "conclusive_document_quality": {
          "properties": {
              # done
              "missing_back": "clear",
              # pedning - part of fraud check/tampering detection - TBD next week
              "digital_document": "clear",
              "punctured_document": "clear",
              "corner_removed": "clear",
              "watermarks_digital_text_overlay": "clear",
              "abnormal_document_features": "clear",
              "obscured_security_features": "clear",
              "obscured_data_points": "clear"
            },
          "result": "clear"
        },
        "image_quality": {
          "properties": {
              "blurred_photo": blurred,
              "covered_photo": missing_fields,
              ## pending
              "cut_off_document": "clear",
              "glare_on_photo": glare,
              "other_photo_issue": missing_fields,
              ## pending
              "two_documents_uploaded": "clear"
          },
          "result": "clear"
        },
        ## pending - complete this by checking excel comments - identifiers for document type
        "supported_document": {
          "properties": {},
          "result": "clear"
        }
      },
      "result": "clear"
    }

    if coloured:
        image_integrity['breakdown']['colour_picture']['result'] = coloured

    if back_id_text and identify_back_id(back_id_text):
        image_integrity['breakdown']['conclusive_document_quality']['properties']['missing_back'] = 'clear'
    else:
        image_integrity['breakdown']['conclusive_document_quality']['properties']['missing_back'] = 'consider'

    f_result = identify_front_id(front_id_text)
    b_result = identify_back_id(back_id_text)

    if f_result and b_result:
        image_integrity['breakdown']['supported_document']['result'] = 'clear'
    else:
        image_integrity['breakdown']['supported_document']['result'] = 'consider'

    # image_integrity['breakdown']['image_quality']['properties']['blurred_photo'] = blurred
    # image_integrity['breakdown']['image_quality']['properties']['glare_on_photo'] = glare

    image_quality_result = create_final_result(image_integrity['breakdown']['image_quality'])
    conclusive_document_quality_result = create_final_result(image_integrity['breakdown']['conclusive_document_quality'])
    colour_picture_result = image_integrity['breakdown']['colour_picture']['result']
    supported_documents_result = image_integrity['breakdown']['supported_document']['result']

    if image_quality_result == 'consider' or conclusive_document_quality_result == 'consider' or colour_picture_result == 'consider' or supported_documents_result == 'consider':
        image_integrity['result'] = 'consider'

    return image_integrity

def visual_authenticity_check(data, front_id_text, back_id_text, facial_similarity):
    visual_authenticity = {
      "breakdown": {
        ## pending - tamper detection
        "digital_tampering": {
          "properties": {},
          "result": "clear"
        },
        "face_detection": {
          "properties": {},
          "result": "clear"
        },
        ## pending - tamper detection
        "fonts": {
          "properties": {},
          "result": "clear"
        },
        "original_document_present": {
          "properties": {
                "scan": "clear",
                ## pending - all 3 - tamper detection
                "document_on_printed_paper": "clear",
                "screenshot": "clear",
                "photo_of_screen": "clear"
          },
          "result": "clear"
        },
        "other": {
          "properties":  {},
          "result": "clear"
        },
        # compare all 3 faces extracted for similarity
        "picture_face_integrity": {
          "properties": {},
          "result": "clear"
        },
        "security_features": {
          "properties": {},
          "result": "clear"
        },
        ## pending - tamper detection
        "template": {
          "properties": {},
          "result": "clear"
        }
      },
      "result": "clear"
    }

    if facial_similarity>=0.65:
        visual_authenticity['breakdown']['face_detection'] = 'clear'
        visual_authenticity['breakdown']['security_features'] = 'clear'
    else:
        visual_authenticity['breakdown']['face_detection'] = 'consider'
        visual_authenticity['breakdown']['security_features'] = 'consider'
    
    doc_type1 = identify_document_type(front_id_text)
    doc_type2 = identify_document_type(back_id_text)
    if doc_type1 == 'EID' or doc_type2=='EID':
        visual_authenticity['breakdown']['original_document_present']['properties']['scan'] = 'clear'
    else:
        visual_authenticity['breakdown']['original_document_present']['properties']['scan'] = 'consider'

    final_result = create_final_result(visual_authenticity)
    visual_authenticity['result'] = final_result

    return visual_authenticity

def main_details(data):
    main_properties = {
        ## barcode is empty in all instance - TBD
        'barcode': [],
        "date_of_birth": "",
        "date_of_expiry": "",
        "document_numbers": [],
        "document_type": "",
        "first_name": "",
        "gender": "",
        "issuing_country": "",
        "last_name": "",
        "mrz_line1": "",
        "mrz_line2": "",
        "mrz_line3": "",
        "nationality": ""
    }

    try:
        main_properties['date_of_birth'] = data.get('dob')
        main_properties['date_of_expiry'] = data.get('expiry_date')

        if data.get('card_number'):
            card_data_t = {
            "type": "type",
            "value": "document_number"
            }

            card_data_v = {
                "type": "value",
                "value": data['card_number']
            }

            main_properties['document_numbers'].append(card_data_t)
            main_properties['document_numbers'].append(card_data_v)

        if data.get('id_number'):
            id_data_t = {
                        "type": "type",
                        "value": "personal_number"
                    }
            id_data_v = {
                        "type": "value",
                        "value": data['id_number']
                    }
                
            main_properties['document_numbers'].append(id_data_t) 
            main_properties['document_numbers'].append(id_data_v) 
        
        main_properties['document_type'] = 'national_identity_card'
        main_properties['first_name'] = data.get('name')
        main_properties['gender'] = data.get('gender')
        main_properties['issuing_country'] = data.get('issuing_place')
        main_properties['last_name'] = data.get('name')
        main_properties['mrz_line1'] = data.get('mrz1')
        main_properties['mrz_line2'] = data.get('mrz2')
        main_properties['mrz_line3'] = data.get('mrz3')
        main_properties['nationality'] = data.get('nationality')

    except:
        main_properties

    return main_properties    

def form_final_data_document_report(data, front_id, front_id_text, back_id, back_id_text, coloured, facial_similarity, blurred, glare, missing_fields):
    try:
        document_report = {
            ## pending - to be filled by dev
            "_id": "",
            "breakdown": {
                "age_validation": age_validation(data.get('dob')),
                "compromised_document": {
                    "result": "clear"
                    },
                "data_comparison": data_comparison_check(),
                "data_consistency": data_consistency_check(data, front_id_text, back_id_text),
                "data_validation": data_validation_check(data),
                "image_integrity": image_integrity_check(front_id_text, back_id_text, coloured, blurred, glare, missing_fields),
                "issuing_authority": {
                "breakdown": {
                    "nfc_active_authentication": {
                    "properties": {}
                    },
                    "nfc_passive_authentication": {
                    "properties": {}
                    }
                }
                },
                "police_record": {},
                "visual_authenticity": visual_authenticity_check(data, front_id_text, back_id_text, facial_similarity),
                ## pending - to be filled by dev
                "check_id": "", 
                "created_at": created_at(),
                "documents": [
                    {
                    ## pending - id value in table stored in db for front id - to be filled by dev
                    "id": ""
                    },
                    {
                    ## pending - id value in table stored in db for front id - to be filled by dev
                    "id": ""
                    }
                ],
                "name": "document",
                "properties": main_details(data),
                # include all results from above to find this
                "result": "",
                "status": "complete",
                "sub_result": ""
            }
        }
        
        final_result = create_final_result(document_report)
        document_report['breakdown']['result'] = final_result
        document_report['breakdown']['sub_result'] = final_result

        return document_report
    
    except Exception as e:
        return e

def form_final_facial_similarity_report(facial_similarity, liveness_result):
    facial_report = {
        "created_at": "",
        ## pending - to be filled by dev
        "href": "/v3.6/reports/<REPORT_ID>",
        ## pending - to be filled by dev
        "id": "<REPORT_ID>",
        "name": "facial_similarity_video",
        "properties": {},
        "breakdown": {
            "face_comparison": {
                "breakdown": {
                    "face_match": { 
                        "properties": {
                            "score": 0,
                            ## pending - to be filled by dev
                            "document_id": "<DOCUMENT_ID>"
                        },
                        "result": "clear", 
                    }
                },
                "result": "clear"
            },
            "image_integrity": {
                "breakdown": {
                    "face_detected": {
                        "result": "clear",
                        "properties": {}
                    },
                    ## pending
                    "source_integrity": {
                        "result": "clear",
                        "properties": {}
                    }
                },
                "result": "clear",
            },
            "visual_authenticity": {
                "breakdown": {
                    "liveness_detected": {
                        "properties": {},
                        "result": "clear",
                    },
                    "spoofing_detection": {
                        "properties": {
                            "score": 0.90
                        },
                        "result": "clear",
                    }
                },
                "result": "clear",
            }
        },
        "result": "clear",
        "status": "complete",
        "sub_result": "clear",
        ## pending - to be filled by dev
        "check_id": "<CHECK_ID>",
        "documents": []
        }
    
    facial_report['created_at'] = created_at()

    facial_report['breakdown']['face_comparison']['breakdown']['face_match']['properties']['score'] = facial_similarity

    if facial_similarity>=0.65:
        facial_report['breakdown']['face_comparison']['breakdown']['face_match']['result'] = 'clear'
        facial_report['breakdown']['face_comparison']['result'] = 'clear'
        facial_report['breakdown']['image_integrity']['breakdown']['face_detected']['result'] = 'clear'
    else:
        facial_report['breakdown']['face_comparison']['breakdown']['face_match']['result']  = 'consider'
        facial_report['breakdown']['face_comparison']['result'] = 'consider'
        facial_report['breakdown']['image_integrity']['breakdown']['face_detected']['result'] = 'consider'
    
    facial_report['breakdown']['visual_authenticity']['breakdown']['spoofing_detection']['properties']['score'] = liveness_result

    if liveness_result == 'consider':
        facial_report['breakdown']['visual_authenticity']['breakdown']['liveness_detected']['result'] = 'consider'
        facial_report['breakdown']['visual_authenticity']['breakdown']['spoofing_detection']['result'] = 'consider'

    visual_authenticity_final_result = create_final_result(facial_report['breakdown']['visual_authenticity'])
    facial_report['breakdown']['visual_authenticity']['result'] = visual_authenticity_final_result

    image_integrity_final_result = create_final_result(facial_report['breakdown']['image_integrity'])
    facial_report['breakdown']['image_integrity']['result'] = image_integrity_final_result

    complete_final_reult = create_final_result(facial_report['breakdown'])
    facial_report['result'] = complete_final_reult
    facial_report['sub_result'] = complete_final_reult

    return facial_report



import base64
import cv2
import io
import numpy as np
import re
from datetime import datetime
from PIL import Image
from skimage.transform import radon
from google.cloud import vision_v1
from pkg_resources import resource_filename
from idvpackage import ocr_utils
import face_recognition
import tempfile

class IdentityVerification:

    def __init__(self):
        """
        This is the initialization function of a class that imports a spoof model and loads an OCR
        reader.
        """
        #self.images = images
        credentials_path = resource_filename('idvpackage', 'streamlit-connection-b1a38b694505.json')
        #credentials_path = "streamlit-connection-b1a38b694505.json"
        self.client = vision_v1.ImageAnnotatorClient.from_service_account_json(credentials_path)
        
    def image_conversion(self,image):  
        """
        This function decodes a base64 string data and returns an image object.
        :return: an Image object that has been created from a base64 encoded string.
        """
        # image=image.split(',')[-1]
        # Decode base64 String Data
        img=Image.open(io.BytesIO(base64.decodebytes(bytes(image, "utf-8"))))
        return img

    def rgb2yuv(self, img):
        """
        Convert an RGB image to YUV format.
        """
        try:
            img=np.array(img)
            return cv2.cvtColor(img, cv2.COLOR_RGB2YUV)
        except Exception as e:
            raise Exception(f"Error: {e}")
    
    def find_bright_areas(self, image, brightness_threshold):
        gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        thresh_image = cv2.threshold(gray_image, brightness_threshold, 255, cv2.THRESH_BINARY)[1]
        contours, hierarchy = cv2.findContours(thresh_image, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

        bright_areas = []

        for contour in contours:
            bounding_box = cv2.boundingRect(contour)

            area = bounding_box[2] * bounding_box[3]

            if area > 800:
                bright_areas.append(bounding_box)

        return len(bright_areas)

    def is_blurry(self, image):
        gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        
        laplacian_variance = cv2.Laplacian(gray_image, cv2.CV_64F).var()
        
        return laplacian_variance
    
    def check_image_quality(self, image, brightness_threshold=210, blur_threshold=250):
        bright_check = False
        blur_check = False

        try:
            # Check if the image can be converted from RGB to YUV
            yuv_img = self.rgb2yuv(self.image_conversion(image))

        except Exception as e:
            raise Exception("Failed to convert image from RGB to YUV: " + str(e))

        try:
            # Check brightness
            brightness = np.average(yuv_img[..., 0])
            if brightness > brightness_threshold:
                raise Exception(f"Image is too bright. Brightness: {brightness}, Threshold: {brightness_threshold}")
            else:
                bright_check = True
        except Exception as e:
            raise Exception("Failed to check image brightness: " + str(e))

        try:
            # Check blurriness
            image = np.array(self.image_conversion(image))
            gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
            fm = cv2.Laplacian(gray, cv2.CV_64F).var()
            if fm < blur_threshold:
                raise Exception(f"Image is too blurry. Blurriness: {fm}, Threshold: {blur_threshold}")
            else:
                blur_check = True
        except Exception as e:
            raise Exception("Failed to check image blurriness: " + str(e))
        
        if bright_check and blur_check:
            return True
    # def check_image_quality(self, id_card, brightness_threshold=245, blur_threshold=150):
    #     id_card = self.image_conversion(id_card)
    #     id_card = np.array(id_card)
    #     bright_result = self.find_bright_areas(id_card, brightness_threshold)
    #     blurry_result = self.is_blurry(id_card)

    #     if bright_result > 1:
    #         raise Exception(f"Image is too bright. Threshold: {brightness_threshold}")

    #     if blurry_result < blur_threshold:
    #         raise Exception(f"Image is too blurry. Blurriness: {blurry_result}, Threshold: {blur_threshold}")

    def process_image(self,front_id):
        img = self.image_conversion(front_id)
        img = np.array(img)
        I = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        h, w = I.shape
        if (w > 640):
            I = cv2.resize(I, (640, int((h / w) * 640)))
        I = I - np.mean(I)
        sinogram = radon(I)
        r = np.array([np.sqrt(np.mean(np.abs(line) ** 2)) for line in sinogram.transpose()])
        rotation = np.argmax(r)
        angle = round(abs(90 - rotation)+0.5)

        if abs(angle) > 5:
            color_img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            im = Image.fromarray(color_img)
            out = im.rotate(angle, expand=True)
        else:
            out = Image.fromarray(img)
    
        # im = self.image_conversion(front_id)
        # out = im.rotate(angle, expand=True)

        return out

    def is_colored(self, base64_image):
        img = self.image_conversion(base64_image)
        img = np.array(img)

        return len(img.shape) == 3 and img.shape[2] >= 3
    
    def get_blurred_and_glared_for_doc(self, front_id, back_id):
        brightness_threshold=210
        blurred = 'clear'
        glare = 'clear'

        front_id_img = self.image_conversion(front_id)
        front_id_arr = np.array(front_id_img)

        back_id_img = self.image_conversion(back_id)
        back_id_arr = np.array(back_id_img)

        blurry1 = self.is_blurry(front_id_arr)
        blurry2 = self.is_blurry(back_id_arr)
        if blurry1 < 150 or blurry2 < 150:
            blurred = 'consider'

        
        yuv_front = self.rgb2yuv(self.image_conversion(front_id))
        yuv_back = self.rgb2yuv(self.image_conversion(back_id))
        brightness1 = np.average(yuv_front[..., 0])
        brightness2 = np.average(yuv_back[..., 0])
        if brightness1 > brightness_threshold or brightness2 > brightness_threshold:
            glare = 'consider'

        # glare1 = self.find_bright_areas(front_id_arr, 245)
        # glare2 = self.find_bright_areas(back_id_arr, 245)
        # if glare1 > 5 or glare2 > 5:
        #     glare = 'consider'
        
        return blurred, glare

    # def get_face_orientation(self, face_landmarks):
    #     left_eye = np.array(face_landmarks['left_eye']).mean(axis=0)
    #     right_eye = np.array(face_landmarks['right_eye']).mean(axis=0)

    #     eye_slope = (right_eye[1] - left_eye[1]) / (right_eye[0] - left_eye[0])
    #     angle = np.degrees(np.arctan(eye_slope))

    #     return angle

    # def extract_selfie_from_video(self, video):
    #     cap = cv2.VideoCapture(video)

    #     # Initialize variables to keep track of the best frame and face score
    #     best_frame = None
    #     best_score = 0

    #     frame_count = 0

    #     while True:
    #         ret, frame = cap.read()
    #         if not ret:
    #             break

    #         frame_count += 1
            
    #         if frame_count % 3 != 0:
    #             continue
                
    #         rgb_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

    #         face_locations = face_recognition.face_locations(rgb_frame)
    #         face_landmarks_list = face_recognition.face_landmarks(rgb_frame)

    #         face_score = len(face_locations)

    #         for landmarks in face_landmarks_list:
    #             angle = self.get_face_orientation(landmarks)
    #             print(f"Current angle: {angle}")
    #             if abs(angle) < 1:
    #                 if face_score > best_score:
    #                     best_score = face_score
    #                     best_frame = frame.copy()


    #     # Release the video capture and close all windows
    #     cap.release()
    #     cv2.destroyAllWindows()

    #     # Save the best frame to a file (optional)
    #     if best_frame is not None:
    #         cv2.imwrite("best_frame.jpg", best_frame)
    #         print("Best frame saved as best_frame.jpg")
    #         return best_frame
    #     else:
    #         print("No suitable frame found.")
    #         return None

    def load_and_process_image_fr(self, base64_image):
        img = self.process_image(base64_image)
        img = np.array(img)
        image = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)

        # base64_image = base64_image.split(',')[-1]
        # image_data = base64.b64decode(base64_image)
        # image_file = io.BytesIO(image_data)

        # image = face_recognition.load_image_file(image_file)

        face_locations = face_recognition.face_locations(image)

        if not face_locations:
            return [], []
    
        face_encodings = face_recognition.face_encodings(image, face_locations)

        return face_locations, face_encodings
    
    def calculate_similarity(self, face_encoding1, face_encoding2):
        similarity_score = 1 - face_recognition.face_distance([face_encoding1], face_encoding2)[0]
        return similarity_score

    def extract_face_and_compute_similarity(self, selfie, front_id):
        face_locations1, face_encodings1 = self.load_and_process_image_fr(selfie)

        face_locations2, face_encodings2 = self.load_and_process_image_fr(front_id)

        if not face_encodings1 or not face_encodings2:
            raise ValueError("No faces detected in one or both images")
        else:
            # face_encoding1 = face_encodings1[0]
            # face_encoding2 = face_encodings2[0]
            largest_face_index1 = face_locations1.index(max(face_locations1, key=lambda loc: (loc[2] - loc[0]) * (loc[3] - loc[1])))
            largest_face_index2 = face_locations2.index(max(face_locations2, key=lambda loc: (loc[2] - loc[0]) * (loc[3] - loc[1])))

            face_encoding1 = face_encodings1[largest_face_index1]
            face_encoding2 = face_encodings2[largest_face_index2]

            similarity_score = self.calculate_similarity(face_encoding1, face_encoding2)

            return similarity_score
    
    def calculate_landmarks_movement(self, current_landmarks, previous_landmarks):
        return sum(
            abs(cur_point.position.x - prev_point.position.x) +
            abs(cur_point.position.y - prev_point.position.y)
            for cur_point, prev_point in zip(current_landmarks, previous_landmarks)
        )

    def calculate_face_movement(self, current_face, previous_face):
        return abs(current_face[0].x - previous_face[0].x) + abs(current_face[0].y - previous_face[0].y)

    def calculate_liveness_result(self, eyebrow_movement, nose_movement, lip_movement, face_movement):
        eyebrow_movement_threshold = 15.0
        nose_movement_threshold = 15.0
        lip_movement_threshold = 15.0
        face_movement_threshold = 10.0

        if (
            eyebrow_movement > eyebrow_movement_threshold or
            nose_movement > nose_movement_threshold or
            lip_movement > lip_movement_threshold or
            face_movement > face_movement_threshold
        ):
            return True
        else:
            return False

    def check_for_liveness(self, video_bytes):
        # cap = cv2.VideoCapture(video)
        with tempfile.NamedTemporaryFile(delete=False) as temp_video_file:
            temp_video_file.write(video_bytes)
        
        cap = cv2.VideoCapture(temp_video_file.name)

        frame_count = 0
        previous_landmarks = None
        previous_face = None
        liveness_result_list = []

        while cap.isOpened():
            ret, frame = cap.read()
            if not ret:
                break
            
            frame_count += 1

            if frame_count % 10 == 0:  # analyze every 10 frames
                _, buffer = cv2.imencode('.jpg', frame)
                image_data = buffer.tobytes()

                image = vision_v1.Image(content=image_data)

                response = self.client.face_detection(image=image)
                faces = response.face_annotations

                largest_face = None
                largest_face_area = 0

                for face in faces:
                    current_landmarks = face.landmarks
                    current_face = face.bounding_poly.vertices
                    face_area = abs((current_face[2].x - current_face[0].x) * (current_face[2].y - current_face[0].y))

                    if face_area > largest_face_area:
                        largest_face = face
                        largest_face_area = face_area

                if largest_face:
                    # face = faces[0]
                    current_landmarks = largest_face.landmarks
                    current_face = largest_face.bounding_poly.vertices

                    if previous_landmarks and previous_face:
                        eyebrow_movement = self.calculate_landmarks_movement(current_landmarks[:10], previous_landmarks[:10])

                        nose_movement = self.calculate_landmarks_movement(current_landmarks[10:20], previous_landmarks[10:20])

                        lip_movement = self.calculate_landmarks_movement(current_landmarks[20:28], previous_landmarks[20:28])

                        face_movement = self.calculate_face_movement(current_face, previous_face)

                        liveness_result = self.calculate_liveness_result(eyebrow_movement, nose_movement, lip_movement, face_movement)

                        liveness_result_list.append(liveness_result)

                    previous_landmarks = current_landmarks
                    previous_face = current_face

        cap.release()

        if any(liveness_result_list):
            liveness_check_result = 'clear'
        else:
            liveness_check_result = 'consider'

        return liveness_check_result

    def extract_id_number_from_front_id(self, front_id_text):
        try:
            id_number_match = re.search(r'.*ID Number\n*([\d-]+)', front_id_text)
            if id_number_match:
                id_number = id_number_match.group(1).replace('-','')
            else:
                id_number = ''
        except:
            id_number = ''

        return id_number
    
    def extract_name_from_front_id(self, front_id_text):
        try:
            name_match = re.search(r'Name: (.+)', front_id_text)
            if name_match:
                name = name_match.group(1)
            else:
                name = ''
        except:
            name = ''

        return name
    
    def extract_dob_from_fron_id(self, front_id_text):
        try:
            date_matches = re.findall(r'(\d{2}/\d{2}/\d{4})', front_id_text)
            date_objects = [datetime.strptime(date, '%d/%m/%Y') for date in date_matches]
            date_of_birth = min(date_objects).strftime('%d/%m/%Y')
        except:
            date_of_birth = ''
        
        return date_of_birth
    
    def extract_expiry_date_from_fron_id(self, front_id_text):
        try:
            date_matches = re.findall(r'(\d{2}/\d{2}/\d{4})', front_id_text)
            date_objects = [datetime.strptime(date, '%d/%m/%Y') for date in date_matches]
            expiry_date = max(date_objects).strftime('%d/%m/%Y')
        except:
            expiry_date = ''
        
        return expiry_date

    def get_ocr_results(self, processed_back_id):
        with io.BytesIO() as output:
            processed_back_id.save(output, format="PNG")
            image_data = output.getvalue()

        image = vision_v1.types.Image(content=image_data)
        response = self.client.text_detection(image=image)
        id_infos = response.text_annotations

        return id_infos

    def extract_ocr_info(self, selfie, video, front_id, back_id, country='UAE'):
        document_report = {}

        is_colored1 = self.is_colored(front_id)
        is_colored2 = self.is_colored(back_id)
        print(f"colured: {is_colored1}, {is_colored2}")
        colour_picture = 'consider'
        if is_colored1 and is_colored2:
            colour_picture = 'clear'

        processed_selfie = self.process_image(selfie)
        processed_front_id = self.process_image(front_id)
        processed_back_id = self.process_image(back_id)
        
        if country=='UAE':
            # extract text from back id
            id_infos= self.get_ocr_results(processed_back_id)
            text = id_infos[0].description

            # extract text from front id
            front_id_text = self.get_ocr_results(processed_front_id)
            front_id_text = front_id_text[0].description
            
            id_number_pattern = r'(?:ILARE|IDARE)\s*([\d\s]+)'
            #card_number_pattern = r'(?:\b|Card Number\s*/\s*رقم البطاقة\s*)\d{9}(?:\b|\s*)'
            card_number_pattern = r'(\b\d{9}\b)|\b\w+(\d{9})\b|Card Number\s*(\d+)|Card Number\s*/\s*رقم البطاقة\s*(\d+)'
            date_pattern = r'(\d{2}/\d{2}/\d{4})'
            #(\d{2}/\d{2}/\d{4}) Date of Birth|\n
            #expiry_date_pattern = r'\n(\d{2}/\d{2}/\d{4})\s*\n'
            gender_pattern = r'Sex: ([A-Z])|Sex ([A-Z])'
            nationality_pattern = r'([A-Z]+)<<'
            # name_pattern = r'([A-Z]+(?:<<[A-Z]+)+(?:<[A-Z]+)+(?:<[A-Z]+))|([A-Z]+(?:<<[A-Z]+)+(?:<[A-Z]+))|([A-Z]+(?:<[A-Z]+)+(?:<<[A-Z]+)+(?:<[A-Z]+)+)'
            name_pattern = r'(.*[A-Za-z]+<[<]+[A-Za-z].*)'
            occupation_pattern = r'Occupation:\s*([\w\s.]+)'
            employer_pattern = r'Employer:\s*([\w\s.]+)'
            issuing_place_pattern = r'Issuing Place:\s*([\w\s.]+)'
            # mrz_pattern = r'(ILARE.*|IDARE.*)'
            mrz_pattern = r'(ILARE.*\n*.*\n*.*\n*.*|IDARE.*\n*.*\n*.*\n*.*)'
            
            try:
                id_number = re.search(id_number_pattern, text)
                id_number = id_number.group(0).replace(" ", "")[15:30]
                if len(id_number)<30:
                    id_number = self.extract_id_number_from_front_id(front_id_text)
            except:
                id_number = ''
            
            try:
                card_number = re.findall(card_number_pattern, text)
                card_number = [c for c in card_number if any(c)]
                if card_number:
                    card_number = "".join(card_number[0])
            except:
                card_number = ''
            
            dob, expiry_date = '', ''
            
            dates = re.findall(date_pattern, text)
            sorted_dates = sorted(dates, key=lambda x: datetime.strptime(x, '%d/%m/%Y'))

            date = [d for d in sorted_dates if any(d)]
            if date:
                try:
                    dob = "".join(date[0])
                except:
                    dob = ''
                try:
                    expiry_date = "".join(date[1])
                except:
                    expiry_date = ''
            
            if not dob:
                dob = self.extract_dob_from_fron_id(front_id_text)
            
            if not expiry_date:
                expiry_date = self.extract_expiry_date_from_fron_id(front_id_text)

            #expiry_date = re.search(expiry_date_pattern, text)
            
            gender = re.findall(gender_pattern, text)
            if gender:
                gender = "".join(gender[0])
            if not gender:
                gender_pattern = r'(?<=\d)[A-Z](?=\d)'
                gender = re.search(gender_pattern, text)
                gender = gender.group(0) if gender else ''
                
            try:
                nationality = re.search(nationality_pattern, text)
                nationality = nationality.group(1)
            except:
                nationality = ''
            
            try:
                name = re.findall(name_pattern, text)
                name = [n for n in name if any(n)]
                if name:
                    name = "".join(name[0])
                    name = name.replace('<',' ').strip()
                if len(name)<5:
                    name = self.extract_name_from_front_id(front_id_text)
            except:
                name = ''
            
            try:
                occupation = re.search(occupation_pattern, text, re.IGNORECASE)
                occupation = occupation.group(1).strip().split('\n', 1)[0]
            except:
                occupation = ''
            
            try:
                employer = re.search(employer_pattern, text, re.IGNORECASE)
                employer = employer.group(1).strip().split('\n', 1)[0]
            except:
                employer = ''
                    
            try:
                issuing_place = re.search(issuing_place_pattern, text, re.IGNORECASE)
                issuing_place = issuing_place.group(1).strip().split('\n', 1)[0]
            except:
                issuing_place = ''
            
            try:
                mrz = re.findall(mrz_pattern, text, re.MULTILINE)
                input_str = mrz[0].replace(" ", "")
                mrz1, remaining = input_str.split("\n", 1)
                mrz2, mrz3 = remaining.rsplit("\n", 1)
            # mrz1, mrz2, mrz3 = mrz[0].replace(' ','')mrz.split("\n")
            except:
                mrz, mrz1, mrz2, mrz3 = '', '', '', ''
            
            similarity = self.extract_face_and_compute_similarity(selfie, front_id)
            similarity = similarity + 0.25

            info_dict = {
                'id_number': id_number,
                'card_number': card_number,
                'name': name,
                'dob': dob ,
                'expiry_date': expiry_date,
                'gender': gender,
                'nationality': nationality,
                'occupation': occupation,
                'employer': employer,
                'issuing_place': issuing_place,
                'mrz': mrz,
                'mrz1': mrz1,
                'mrz2': mrz2,
                'mrz3': mrz3,
                'similarity': similarity
            }
            
            blurred, glare = self.get_blurred_and_glared_for_doc(front_id, back_id)
            missing_fields = 'clear'

            non_optional_keys = ["id_number", "card_number", "name", "dob", "expiry_date", "gender", "nationality", "mrz", "mrz1", "mrz2", "mrz3", "similarity"]
            empty_string_keys = [key for key, value in info_dict.items() if key not in non_optional_keys and value == '']

            if empty_string_keys:
                missing_fields = 'consider'
            
            document_report = ocr_utils.form_final_data_document_report(info_dict, front_id, front_id_text, back_id, text, colour_picture, similarity, blurred, glare, missing_fields)
            
            liveness_result = self.check_for_liveness(video)
            facial_report = ocr_utils.form_final_facial_similarity_report(similarity, liveness_result)

        else:
            pass
        
        #json_object = json.dumps(df, indent = 4) 
        return info_dict, document_report, facial_report
